#include "Common.h"

namespace CryptoPP {
	typedef ::CryptoPP::RSASS<::CryptoPP::PKCS1v15, ::CryptoPP::SHA256>::Verifier RSASSA_PKCS1v15_SHA256_Verifier;
	typedef ::CryptoPP::RSASS<::CryptoPP::PKCS1v15, ::CryptoPP::SHA256>::Signer RSASSA_PKCS1v15_SHA256_Signer;
}


// convert UTF-8 string to wstring
std::wstring utf8_to_wstring(const std::string& str)
{
	std::wstring_convert<std::codecvt_utf8<wchar_t>> myconv;
	return myconv.from_bytes(str);
}

// convert wstring to UTF-8 string
std::string wstring_to_utf8(const std::wstring& str)
{
	std::wstring_convert<std::codecvt_utf8<wchar_t>> myconv;
	return myconv.to_bytes(str);
}

// May be needed.
bool LoadRSAFromString(std::string &str, CryptoPP::AutoSeededRandomPool &rng, CryptoPP::RSA::PrivateKey & key)
{
	string RSA_PRIV_KEY = str;

	static string HEADER = "-----BEGIN RSA PRIVATE KEY-----";
	static string FOOTER = "-----END RSA PRIVATE KEY-----";

	size_t pos1, pos2;
	pos1 = RSA_PRIV_KEY.find(HEADER);
	if (pos1 == string::npos)
	{
		GenerateError2(L"PEM header not found");
		return false;
	}

	pos2 = RSA_PRIV_KEY.find(FOOTER, pos1 + 1);
	if (pos2 == string::npos)
	{
		GenerateError2(L"PEM footer not found");
		return false;
	}


	// Start position and length
	pos1 = pos1 + HEADER.length();
	pos2 = pos2 - pos1;
	string keystr = RSA_PRIV_KEY.substr(pos1, pos2);

	// Base64 decode, place in a ByteQueue    
	CryptoPP::ByteQueue queue;
	CryptoPP::Base64Decoder decoder;

	decoder.Attach(new CryptoPP::Redirector(queue));
	decoder.Put((const byte*)keystr.data(), keystr.length());
	decoder.MessageEnd();

	try
	{
		key.BERDecodePrivateKey(queue, false /*paramsPresent*/, queue.MaxRetrievable());

		// BERDecodePrivateKey is a void function. Here's the only check
		// we have regarding the DER bytes consumed.
		if (!queue.IsEmpty())
		{
			GenerateError2(L"Didn't read all the bytes of RSA, which is confusing.");
			return false;
		}

		bool valid = key.Validate(rng, 3);
		if (!valid)
		{
			GenerateError2(L"RSA private key is not valid");
			return false;
		}

		std::stringstream st;
		st << "N:" << key.GetModulus() << endl
		   << "E:" << key.GetPublicExponent() << endl
		   << "D:" << key.GetPrivateExponent() << endl;
		OutputDebugStringA(st.str().c_str());

		return true;
	}
	catch (const std::exception & ex)
	{
		std::wostringstream err;
		err << L"Something threw an exception while reading:\r\n" << utf8_to_wstring(ex.what());
		GenerateError2(err);
		return false;
	}
}


#define GenerateError2(x)
void ServiceLogin(const wchar_t * file, const wchar_t *  levelOfControl_, const wchar_t * keyFile_, int useBase64Decoder_, int stripHeaders_, int useLoad_)
{
	// Perms taken from 
	// https://cloud.google.com/storage/docs/authentication?hl=en
	//
	wstring perms, levelOfControl(levelOfControl_), keyFile(keyFile_);
	if (levelOfControl == L"full-control")
		perms = L"https://www.googleapis.com/auth/devstorage.full_control";
	else if (levelOfControl == L"read-only")
		perms = L"https://www.googleapis.com/auth/devstorage.read_only";
	else if (levelOfControl == L"read-write")
		perms = L"https://www.googleapis.com/auth/devstorage.read_write";
	else
	{
		GenerateError2(L"Using unrecognised perm in Cake.");
		return;
	}
	
	std::wifstream wif(file);
	if (wif.fail())
	{
		GenerateError2(L"Opening JSON file failed.");
		return;
	}
	wif.imbue(std::locale(std::locale::empty(), new std::codecvt_utf8<wchar_t>));
	wstring state;

	utf8string msgUTF8 = "", b64jwtSig, jwtClaimUTF8;
	auto json = web::json::value::parse(wif);
	wstring clientID = json[L"client_id"].as_string();
	wstring	clientSecret = L"";// json[L"client_secret"].as_string();
	wstring auth_url = json[L"auth_uri"].as_string();
	wstring token_uri = json[L"token_uri"].as_string();
	wstring private_key = json[L"private_key"].as_string();
	string private_keyC = wstring_to_utf8(private_key);

	time_t curTimeUTCSeconds = { 0 };
	time(&curTimeUTCSeconds);

	// See https://developers.google.com/identity/protocols/OAuth2ServiceAccount
	std::ostringstream jwt;
	
	// First, append JWT header: {"alg":"RS256","typ":"JWT"}
	jwt << "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9" << ".";

	// Second, append JWT claim set
	auto jwtClaimJSON = web::json::value::object(
	{
		pair<string_t, value>(wstring(L"iis"), value(L"ctfdvrext@phi-fusion-tests.iam.gserviceaccount.com", true)),
		pair<string_t, value>(wstring(L"aud"), value(L"https://www.googleapis.com/oauth2/v4/token", true)),
		pair<string_t, value>(wstring(L"scope"), value(perms, true)),
		pair<string_t, value>(wstring(L"iat"), value((__int64)curTimeUTCSeconds)),
		pair<string_t, value>(wstring(L"exp"), value((__int64)curTimeUTCSeconds + 3600LL)) // Max limit is 1 hr, 60m * 60s
	});
	jwtClaimUTF8 = wstring_to_utf8(jwtClaimJSON.serialize());

	// If message is good, encode it to B64
	utf8string b64jwtHeader = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9", b64jwtClaim;
	CryptoPP::StringSource ss4(jwtClaimUTF8, true, new CryptoPP::Base64URLEncoder(new CryptoPP::StringSink(b64jwtClaim)));
	msgUTF8 = b64jwtHeader + "." + b64jwtClaim;
	int index;
	if (useLoad_ > 0)
	{
		index = b64jwtClaim.find_first_of('=');
		if (index != std::string::npos)
			b64jwtClaim = b64jwtClaim.substr(0U, index);
	}

	// TODO: Dot may be appeneded as message part
	// Third, use this text:
	// 
	// Sign it with RSASSA-PKCS1-V1_5-SIGN, using a private key obtained through Google Developer Console.
	// The output will be a byte array.
	// With that output, serialise it to base64url. This is the 3rd part.
	// RSASSA_PKCS1v15_SHA256_Signer
	// Generate or Load keys
	//Read private key
	CryptoPP::ByteQueue bytesPriv;

	CryptoPP::RSA::PrivateKey privateKey;
	// Base64 decode, place in a ByteQueue    
	CryptoPP::Base64URLEncoder encoder;/*
	CryptoPP::FileSource fs(keyFile.c_str(), true);
	
	std::string moo;
	CryptoPP::StringSink str(moo);
	fs.TransferTo(str);
	if (moo.empty())
	{
		std::wostringstream st;
		st << L"Key file path not found:\r\n" << keyFile;
		GenerateError2(st);
		return;
	}
	CryptoPP::AutoSeededRandomPool rng;
	if (!LoadRSAFromString(moo, rng, privateKey))
		return;
		*/

	CryptoPP::AutoSeededRandomPool rng;
	std::string header = "-----BEGIN PRIVATE KEY-----\n";
	std::string privateKeyNoHeadersB64Str = private_keyC.substr(header.size()), privateKeyNoHeadersBINStr;
	privateKeyNoHeadersB64Str = privateKeyNoHeadersB64Str.substr(0, privateKeyNoHeadersB64Str.find_last_of("-="));
	CryptoPP::StringSource st(privateKeyNoHeadersB64Str, true,
		new CryptoPP::Base64URLDecoder(
		new CryptoPP::StringSink(privateKeyNoHeadersBINStr)));
	
	CryptoPP::StringSource privateKeyNoHeadersBINSrc(privateKeyNoHeadersBINStr, true);
	CryptoPP::StringSource privateKeyHeadersSrc(private_keyC, true);
	CryptoPP::StringSource privateKeyNoHeadersB64Src(privateKeyNoHeadersB64Str, true);
	//privateKey.Load(privateKeyNoHeadersBINSrc);
	privateKey.BERDecodePrivateKey(privateKeyNoHeadersBINSrc, false, privateKeyNoHeadersBINSrc.MaxRetrievable());

	////////////////////////////////////////////////
	// Sign and Encode
	CryptoPP::RSASSA_PKCS1v15_SHA256_Signer signer(privateKey);

	state = L"Cake - random gen";

	CryptoPP::RSASSA_PKCS1v15_SHA256_Verifier publicKey(privateKey);
	CryptoPP::StringSource ss3(msgUTF8, true,
		new CryptoPP::SignerFilter(rng, signer,
		new CryptoPP::Base64URLEncoder(
		new CryptoPP::StringSink(b64jwtSig))));
	// signature contains signed text signature
	
	if (useLoad_ > 0)
	{
		index = b64jwtSig.find_first_of('=');
		if (index != std::string::npos)
			b64jwtSig = b64jwtSig.substr(0U, b64jwtSig.size() - 2);
	}
	////////////////////////////////////////////////
	// Verify and Recover
	//CryptoPP::RSASSA_PKCS1v15_SHA256_Verifier verifier(publicKey);
	/*
	CryptoPP::StringSource ss2(msgUTF8 + signature, true,
	new CryptoPP::SignatureVerificationFilter(
	verifier, NULL,
	CryptoPP::SignatureVerificationFilter::THROW_EXCEPTION
	) // SignatureVerificationFilter
	); // StringSource
	*/
	/*
	catch (const std::exception & exc)
	{
		std::wostringstream ws;
		ws << L"Error of type:\r\n" << utf8_to_wstring(exc.what()) << L"\r\nState:\r\n" << state;
		GenerateError2(ws);
		return;
	}*/

	string jwtFinal = b64jwtHeader + "." + b64jwtClaim + "." + b64jwtSig;
	wstring jwtFinalWS = utf8_to_wstring(jwtFinal);
	std::stringstream jwtFinalSS; jwtFinalSS << jwtFinal;
	MessageBoxA(NULL, jwtFinal.c_str(), "Final JWS listing", MB_OK);

	auto jsonBody = web::json::value::object(
	{
		pair<string_t, value>(wstring(L"grant_type"), value(L"urn:ietf:params:oauth:grant-type:jwt-bearer", true)),
		pair<string_t, value>(wstring(L"assertion"), value(jwtFinalWS, true))
	});
	
	//wstring s = oauth2Config->build_authorization_uri(true);

	auto h = make_shared<http_client>(L"https://www.googleapis.com/oauth2/v4/");
	auto h2 = make_shared<http_client>(L"http://www.googleapis.com/oauth2/v4/");

	h2->request(methods::POST, L"/token", L"grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Ajwt-bearer&assertion=" + jwtFinalWS, L"application/x-www-form-urlencoded"//, L"",
		//L"application/x-www-form-urlencoded"
		).wait();

	h->request(methods::POST, L"/token", L"grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Ajwt-bearer&assertion=" + jwtFinalWS, L"application/x-www-form-urlencoded"//, L"",
		//L"application/x-www-form-urlencoded"
		).then([&](http_response response)
	{
		return response.content_ready();
	}).then([&](http_response response)
	{
		if (response.status_code() != status_codes::OK)
		{
			std::wostringstream str;
			str << response.extract_string().get();
			/*
			Concurrency::streams::stringstreambuf buffer;
			std::wstring fred3;
			response.body().streambuf() .extract();

			if (buffer.exception() != nullptr)
				fred3 = L"some exception";
			else
			{
				std::string fred5 = buffer.get();
				char * fred = (char *)calloc(buffer.size() + 1U, 1U);
				std::string fred2;
				if (buffer.scopy(fred, _msize(fred)) != _msize(fred))
				{
					fred3 = L"Wot.";
					goto meh;
				}

				fred2 = fred;
				fred3 = utf8_to_wstring(fred2);
				buffer.close();
				meh:
				free(fred);
			}
			*/
			std::wstring fred3 = str.str();
			std::wostringstream ws;
			ws << L"Failed to refresh project list. Error status " << response.status_code() << "\r\n"
				<< L"Reason phrase:\r\n"
				<< response.reason_phrase() << L"\r\n"
				<< L"Error body:\r\n"
				<< fred3;

			GenerateError2(ws);
			return task_from_result(value());
		}


		
		return response.extract_json();
	}).then([&](task<value> previousTask)
	{
		bool enteredCriticalSection = false;
		try
		{
			value js = previousTask.get();
			if (js.is_null())
				return;

			EnterCriticalSection(&lock); enteredCriticalSection = true;

			std::wostringstream ws;
			ws << L"Good to go. Response: \r\n" << js;

			MessageBoxW(NULL, ws.str().c_str(), L"!", MB_OK);
			//storage = make_shared<http_client>(L"https://www.googleapis.com/storage/v1/", *config);
			//bigquery = make_shared<http_client>(L"https://www.googleapis.com/bigquery/v2/", *config);
			//GenerateError2(L"Total items footer does not match content's decoded array size.");
		}
		catch (http_exception exc)
		{
			if (enteredCriticalSection) LeaveCriticalSection(&lock);
			std::wostringstream ws;
			ws << "Got a http_exception:\r\n" << exc.what();
			GenerateError2(ws);
		}
		catch (json_exception exc)
		{
			if (enteredCriticalSection) LeaveCriticalSection(&lock);
			std::wostringstream ws;
			ws << "Got a json_exception:\r\n" << exc.what();
			GenerateError2(ws);
		}

	});
	

	//auto j = o->bearer_auth();
	auto p = oauth2::experimental::oauth2_token(L"!");

	// https://casablanca.codeplex.com/wikipage?title=OAuth%20Client

	//	config->set_oauth2()
	//	config->set_credentials(creds);
}